﻿#!/usr/bin/env python
# -*- coding: utf-8 -*- 

import ctypes
from ConfigParser import SafeConfigParser, NoOptionError
from StringIO import StringIO
from Tkinter import *
from tkMessageBox import showinfo

import pyHook
from tkSnack import initializeSnack, Sound

__version__ = "0.3.4"

user32 = ctypes.windll.user32
WIDTH = user32.GetSystemMetrics(0)  # main monitor width
HEIGHT = user32.GetSystemMetrics(1)  # main monitor height
VWIDTH = user32.GetSystemMetrics(78)  # virtual monitor width
VHEIGHT = user32.GetSystemMetrics(79)  # virtual monitor height
NBSCREENS = user32.GetSystemMetrics(80)
LARG = 200  # côté de la fenêtre principale
OFF = 50    # distance aux bords droit et supérieur


class scrabble_timer(Tk):
    def __init__(self,parent):
        Tk.__init__(self,parent)
        #-- config manager
        self.configfile = 'minuteur.config'
        self.config = SafeConfigParser()
        self.config.readfp(self.getDefaultConfig())  # --> charges d'abord les valeurs par défaut codées en dur
        if not self.config.read(self.configfile):             # --> récrit le fichier de configuration
            with open(self.configfile, 'wb') as configfile:   #            si celui-ci est introuvable
                self.config.write(configfile)
        #-- root manager
        self.withdraw()
        self.iconbitmap('res/appicon.ico')
        self.attributes('-topmost', True)
        self.protocol("WM_DELETE_WINDOW", self.onDestroy)
        #-- clone manager
        clone = self.clone = Toplevel(self)
        clone.withdraw()
        clone.attributes('-topmost', True)
        clone.update_idletasks()
        clone.overrideredirect(True)
        #-- sounds manager
        initializeSnack(self)
        self.tk.eval('package require snackogg')
        self.bip1 = Sound(load='res/bip1.ogg')
        self.bip2 = Sound(load='res/bip2.ogg')
        #-- varia
        self.top = None
        #-- appearence
        self.setColours()
        self.ratio = self.config.getfloat('timer','ratio')
        self.calibration = int(self.config.getfloat('timer','calibration'))
        #-- hooker
        self.hooker()
        self.hwnd = self.winfo_id()
        #-- draw timer
        self.drawtimer()


    def setColours(self):
        u"""Vérifie que le thème existe, qu'il s'agit bien de couleurs valides et qu'il y en a au moins 6"""

        theme = self.config.get('timer','thème')
        try:
            colours = self.config.get('thèmes',theme).split(',')
            for n in range(6):
                self.winfo_rgb(colours[n])
        except (IndexError, TclError, NoOptionError):
            colours = self.config.get('thèmes','défaut').split(',')
        self.timer_bg = colours[0].strip()
        self.ring_on = colours[1].strip()
        self.ring_off = colours[2].strip()
        self.display_bip1 = colours[3].strip()
        self.display_bip2 = colours[4].strip()
        self.display_off = colours[5].strip()


    def drawtimer(self):
        u"""Dessine le minuteur à l'échelle spécifiée par le paramètre ratio."""
        
        r = self.ratio
        self.x, self.y, self.rad = LARG/2*r, LARG/2*r, LARG/8*3*r
        # installation du canvas
        self.grid()
        if self.config.get('timer','mode') == 'big':
            self.canvas = Canvas(self, width=LARG*r, height=LARG*r, bg=self.timer_bg, highlightbackground='darkgray')
        else:
            self.canvas = Canvas(self, width=LARG/1.2*r, height=LARG/2.2*r, bg=self.timer_bg, highlightbackground='darkgray')
        self.canvas.grid(column=0, row=0, sticky="NSEW")
        # enregistrement des événements
        self.bind('<Button-1>',self.onMousePressed)
        self.bind('<B1-Motion>',self.onMouseMotion)
        self.bind("<Double-Button-1>", self.onMouseClick)
        self.bind("<Button-2>", self.onMouseClick)
        self.bind("<Button-3>", self.onMouseClick)
        self.bind("<space>", self.onPressKey)
        self.bind("m", self.onPressKey)
        self.bind("M", self.onPressKey)
        self.bind("q", self.onPressKey)
        self.bind("Q", self.onPressKey)
        self.bind("x", self.onPressKey)
        self.bind("X", self.onPressKey)
        self.bind("/", self.onPressKey)
        self.bind_all("<F1>", self.onPressKey)
        self.bind_all("<F2>", self.onPressKey)
        self.bind_all("<F3>", self.onPressKey)
        self.bind_all("c", self.onPressKey)
        self.bind_all("C", self.onPressKey)
        self.bind_all("d", self.onPressKey)
        self.bind_all("D", self.onPressKey)
        # création du clone
        clone = self.clone
        clone.canvas = Canvas(clone, width=LARG/1.2*r, height=LARG/2.2*r, bg=self.timer_bg, highlightbackground='darkgray')
        clone.canvas.pack()
        clone.bind('<Button-1>',self.onMousePressed)
        clone.bind('<B1-Motion>',self.onMouseMotion)
        clone.geometry( '+{:d}+{:d}'.format(VWIDTH-int(clone.canvas['width'])-20, 10))
        # positionnement et mise en forme générale
        self.geometry( '-{:d}+{:d}'.format(OFF, OFF))
        self.resizable(False, False)
        self.resettimer()
        self.deiconify()
        # doit se trouver après self.deiconify() pour un problème de focus
#        self.update_idletasks()
#        self.overrideredirect(True)


    def resettimer(self):
        u"""Isole les instructions d'affichage qui doivent suivre chaque appel à stoptimer()."""
        
        self.isRunning = None
        self.bip2.stop()
        self.color = self.display_bip1
        x, y, rad, r = self.x, self.y, self.rad, self.ratio
        self.canvas.delete(ALL)
        self.clone.canvas.delete(ALL)
        self.arc = 360
        self.time = map(int, self.config.get('timer','bips').split('+'))
        self.t = sum(self.time)
        self.arc_incr = 360/float(self.t)
        txt = u"{:0>2}:{:0>2}".format(*divmod(self.t, 60))
        if self.config.get('timer','mode') == 'big':
            self.canvas.create_arc(x-rad, y-rad, x+rad, y+rad, width=LARG/16*r, style=ARC,extent=359.999, outline=self.ring_off)
            self.canvas.create_arc(x-rad, y-rad, x+rad, y+rad, width=LARG/16*r, style=ARC,extent=self.arc-.001, start=90, outline=self.ring_on)
            self.canvas.create_text(x, y, fill=self.display_off, text=txt, font=("Helvetica", int(LARG/7*r), 'bold'), tags='display')
        else:
            self.canvas.create_text(x/1.2+2, y/2.2+2, fill=self.display_off, text=txt, font=("Helvetica", int(LARG/4.8*r), 'bold'), tags='display')
        self.clone.canvas.create_text(x/1.2+2, y/2.2+2, fill=self.display_off, text=txt, font=("Helvetica", int(LARG/4.8*r), 'bold'), tags='display')
        self.title( u'{} sec.'.format(self.config.get('timer','bips')) )
        self.clone.lift()
        self.canvas.focus_set()


    def starttimer(self):
        u"""
        Enclenche le décompte jusqu'à t == 0.
        La méthode pausetimer() interrompt le décompte.
        La méthode stoptimer() remet le minuteur à zéro.
        """
        x, y, rad, r = self.x, self.y, self.rad, self.ratio
        self.canvas.delete(ALL)
        self.clone
        self.clone.canvas.delete(ALL)
        self.arc -= self.arc_incr
        self.t -= 1
        if self.t <= self.time[1]:
            self.color = self.display_bip2
            if self.t > self.time[1]-2:  # jouer deux passes
                self.bip1.play()
        else:
            self.color = self.display_bip1
        txt = u"{:0>2}:{:0>2}".format(*divmod(self.t, 60))
        if self.config.get('timer','mode') == 'big':
            self.canvas.create_arc(x-rad, y-rad, x+rad, y+rad, width=LARG/16*r, style=ARC, extent=359.999, outline=self.ring_off, tags='ring_off')
            self.canvas.create_arc(x-rad, y-rad, x+self.rad, y+rad, width=LARG/16*r, style=ARC, extent=self.arc, start=90, outline=self.ring_on, tags='ring_on')
            self.canvas.create_text(x, y, fill=self.color, text=txt, font=("Helvetica", int(LARG/7*r), 'bold'), tags='display')
        else:
            self.canvas.create_text(x/1.2+2, y/2.2+2, fill=self.color, text=txt, font=("Helvetica", int(LARG/4.8*r), 'bold'), tags='display')
        self.clone.canvas.create_text(x/1.2+2, y/2.2+2, fill=self.color, text=txt, font=("Helvetica", int(LARG/4.8*r), 'bold'), tags='display')
        self.clone.lift()
        if self.t == 0:
            bip2_ms = self.bip2.length(unit="SECONDS")*1000
            self.bip2.play()
            self.isRunning = self.canvas.after(bip2_ms, self.resettimer)
        else:
            self.isRunning = self.canvas.after(1000+self.calibration, self.starttimer)


    def pausetimer(self):
        self.isRunning = self.canvas.after_cancel(self.isRunning)
        self.canvas.itemconfig('display',fill=self.display_off)
        self.clone.canvas.itemconfig('display',fill=self.display_off)


    def stoptimer(self):
        if self.isRunning:
            self.isRunning = self.canvas.after_cancel(self.isRunning)
        self.resettimer()


    def changeMode(self):
        if self.config.get('timer','mode') == 'big':
            self.config.set('timer','mode','small')
            self.canvas["height"] = int(self.canvas["height"])/2.2
            self.canvas["width"] = int(self.canvas["width"])/1.2
        else:
            self.config.set('timer','mode','big')
            self.canvas["height"] = int(self.canvas["height"])*2.2
            self.canvas["width"] = int(self.canvas["width"])*1.2

        self.geometry( '-{:d}+{:d}'.format(OFF, OFF)) 
        if not self.isRunning:
            self.resettimer()


    def showhelp(self):
        helptxt = u"""
    -------------------------------------
    Commande globale (hors focus)
    -------------------------------------
    <Oem_7> : lancer | pauser | interrompre la sonnerie finale

    (la touche <Oem_7> est sous <Échappe/Escape>, elle est
     en général dédiée au 2 exposant <²> dans les claviers francophones)  

    --------------------------------
    Commandes clavier
    --------------------------------    {}
    --------------------------------
    Commandes souris
    --------------------------------    {}
    """.format(self.onPressKey.__doc__, self.onMouseClick.__doc__)

        if self.top:
            self.top.destroy()
        top = self.top = Toplevel(self)
        top.title(u"Aide")
        top.bind_all('<Escape>', lambda ev: top.destroy())
        top.bind_all('<Return>', lambda ev: top.destroy())
        Label(top,text=helptxt,justify=LEFT,padx=5).pack(fill=BOTH)
        top.geometry( '-{}+{}'.format(OFF+20, OFF+20) )
        top.resizable(False, False)
        top.attributes('-topmost', True)


    def showparameters(self):
        title = ''
        s = self.s = StringVar()
        if self.top:
            try:
                title = self.top.title()
            except TclError:
                pass
            self.top.destroy()
        # Alterner entre temps de réflexion prédéfinis et temps de réflexion personnel :
        # - ouvre la fenêtre temps persos si la fenêtre temps prédéfinis est ouverte, et inversement
        if title == u"Temps de réflexion":
            def setCustomDelay(event):
                if s1.get() > 0 and s2.get() >= 0:
                    s.set( '{}+{}'.format(s1.get(), s2.get()) )
                    self.setConfig(True)
            top = self.top = Toplevel(self)
            top.grid()
            top.title(u'Temps personnel')
            top.bind('<Escape>', lambda ev: top.destroy())
            top.bind('<Return>', setCustomDelay)
            Label(top, text=u"Temps de réflexion en secondes :").grid(column=0, row=0, columnspan=3, padx=10, pady=10)
            Label(top, text=u"+").grid(column=1, row=1)

            s1 = IntVar()
            s2 = IntVar()
            bips = self.config.get('timer','bips').split('+')
            spin1 = Spinbox(top, from_=10, to=1000, width=4, textvariable=s1, justify=RIGHT, increment=10)
            spin1.delete(0,END)
            spin1.insert(0,bips[0])
            spin1.grid(column=0, row=1, pady=10, sticky='E')
            spin2 = Spinbox(top, from_=10, to=1000, width=4, textvariable=s2, justify=RIGHT, increment=10)
            spin2.delete(0,END)
            spin2.insert(0,bips[1])
            spin2.grid(column=2, row=1, pady=10, sticky='W')
        else:
            top = self.top = Toplevel(self)
            top.title(u"Temps de réflexion")
            top.bind('<Escape>', lambda ev: top.destroy())
            top.bind('<Return>', lambda ev: top.destroy())
            top.bind('<Up>', lambda ev: self.browseoptions(ev, 'presets'))
            top.bind('<Down>', lambda ev: self.browseoptions(ev, 'presets'))
            i = 0
            for p in self.config.items('presets'):
                txt = u" {}   [{} + {} sec.] ".format(p[0].upper(), *p[1].split('+'))
                r = Radiobutton(top, height=2, padx=15, text=txt, variable=s, value=p[1], indicatoron=False, command=self.setConfig)
                r.pack(fill=BOTH)
                if p[1] == self.config.get('timer','bips'):
                    r.select()
                i += 1

        top.geometry( '-{}+{}'.format(LARG+OFF, OFF+20) )
        top.resizable(False, False)
        top.attributes('-topmost', True)
        try:
            spin1.focus_set()
            spin1.selection('range',0,END)
        except UnboundLocalError:
            top.focus_set()


    def showthemes(self):
        if self.top:
            self.top.destroy()
        top = self.top = Toplevel(self)
        top.title(u"Thèmes")
        top.bind('<Escape>', lambda ev: top.destroy())
        top.bind('<Return>', lambda ev: top.destroy())
        top.bind('<Up>', lambda ev: self.browseoptions(ev, 'thèmes'))
        top.bind('<Down>', lambda ev: self.browseoptions(ev, 'thèmes'))
        s = self.s = StringVar()
        i = 0
        for p in self.config.items('thèmes'):
            txt = u'{:^25}'.format(p[0].decode('utf-8').upper())
            r = Radiobutton(top, height=2, padx=15, text=txt, variable=s, value=p[0], indicatoron=False, command=self.setConfig)
            r.pack(fill=BOTH)
            if p[0] == self.config.get('timer','thème'):
                r.select()
            i += 1
        top.geometry( '-{}+{}'.format(LARG+OFF, OFF+20) )
        top.resizable(False, False)
        top.attributes('-topmost', True)
        top.focus_set()


    def browseoptions(self, event, categorie):
        if categorie == "presets":
            opt = [p[1] for p in self.config.items(categorie)]
        else:
            opt = [p[0] for p in self.config.items(categorie)]
        try:
            sel = opt.index(self.s.get().encode('utf-8'))
        except ValueError:
            sel = -1
        if event.keysym == 'Up':
            if sel < 1:
                sel = len(opt)
            self.s.set(opt[sel-1])
        else:
            if sel > len(opt)-2:
                sel = -1
            self.s.set(opt[sel+1])
        self.setConfig(False)
        self.top.focus_set()


    def setConfig(self, closeTop=True):
        if closeTop: self.top.destroy()
        if '+' in self.s.get():     # on met à jour les sonneries
            self.config.set('timer', 'bips', self.s.get())
            if not self.isRunning and self.t == sum(self.time):
                self.time = map(int, self.config.get('timer','bips').split('+'))
                self.resettimer()
        else:  # on met à jour le thème
            self.config.set('timer', 'thème', self.s.get().encode('utf-8'))
            self.setColours()
            self.canvas.config(bg=self.timer_bg)
            self.clone.canvas.config(bg=self.timer_bg)
            if not self.isRunning:
                if self.t == sum(self.time):
                    self.resettimer()
                else:
                    self.canvas.itemconfig('ring_off',outline=self.ring_off)
                    self.canvas.itemconfig('ring_on',outline=self.ring_on)
                    self.canvas.itemconfig('display',fill=self.display_off)
                    self.clone.canvas.itemconfig('display',fill=self.display_off)


    def onMouseClick(self, event):
        u"""
    <Double-clic> : lancer | pauser | interrompre la sonnerie finale
    <Clic central> : stopper | interrompre la sonnerie finale
    <Clic droit> : choisir parmi les temps de réflexion prédéfinis
    <Double clic droit> : définir un temps de réflexion personnel
        """

        if event.num == 1:  # clic gauche
            if self.t == 0:
                self.stoptimer()
            elif self.isRunning:
                self.pausetimer()
            else:
                self.canvas.itemconfig('display',fill=self.color)
                self.clone.canvas.itemconfig('display',fill=self.color)
                self.isRunning = self.canvas.after(1000+self.calibration, self.starttimer)
        elif event.num == 2:  # clic centre
            self.stoptimer()
        elif event.num == 3:  # clic droit
            self.showparameters()


    def onPressKey(self, event):
        u"""
    <Espace> : lancer | pauser | interrompre la sonnerie finale
    <c | C> : si un deuxième écran est connecté, afficher/masquer
                    le clone "tableau"
    <d | D> : afficher/masquer le décorateur de fenêtre
    <m | M> : permuter le mode d'affichage du chrono principal
    <q | Q> : fermer l'application
    <x | X> : stopper | interrompre la sonnerie finale
    <F1> : afficher l'aide
    <F2> : choisir parmi les temps de réflexion prédéfinis
    2x <F2> : personnaliser le temps de réflexion
    <F3> : choisir un thème
        """

        if event.keysym == "space":
            if self.t == 0:
                self.stoptimer()
            elif self.isRunning:
                self.pausetimer()
            else:
                self.canvas.itemconfig('display',fill=self.color)
                self.clone.canvas.itemconfig('display',fill=self.color)
                self.isRunning = self.canvas.after(1000+self.calibration, self.starttimer)
        elif event.keysym == "F1":
            self.showhelp()
        elif event.keysym == "F2":
            self.showparameters()
        elif event.keysym == "F3":
            self.showthemes()
        elif event.char == '/':
            if self.isRunning:
                self.stoptimer()
            else:
                self.canvas.itemconfig('display',fill=self.color)
                self.clone.canvas.itemconfig('display',fill=self.color)
                self.isRunning = self.canvas.after(1000+self.calibration, self.starttimer)
        elif event.char in 'cC':
            if NBSCREENS > 1:
                if self.clone.state() == 'withdrawn':
                    self.clone.deiconify()
                else:
                    self.clone.withdraw()
            else:
                showinfo('Avertissement',"Veuillez d'abord brancher un second moniteur.")
        elif event.char in 'dD':  # test !
            self.update_idletasks()
            self.overrideredirect(not(self.overrideredirect()))
        elif event.char in 'mM':
            self.changeMode()
        elif event.char in 'qQ':
            self.onDestroy()
        elif event.char in 'xX':
            self.stoptimer()


    def onMouseMotion(self, ev):
        ev.widget.master.geometry('+{:d}+{:d}'.format(ev.x_root-self.curx,ev.y_root-self.cury))


    def onMousePressed(self, ev):
        self.curx, self.cury= ev.x, ev.y


    def onDestroy(self):
        with open(self.configfile, 'wb') as configfile:
            self.config.write(configfile)
        self.bip1.stop()
        self.bip2.stop()
        self.quit()
        self.destroy()


    def getDefaultConfig(self):
        u"""Renvoie la configuration par défaut"""
        return StringIO(
            "[timer]\n"
            "mode = big\n"
            "bips = 150+30\n"
            "thème = défaut\n"
            "ratio = 0.7\n"
            "calibration = -15\n"
            "\n[presets]\n"
            "normal = 10+30\n"
            "semi-rapide = 15+15\n"
            "blitz = 10+10\n"
            "[thèmes]\n"
            "défaut = lightyellow,light sea green,lightgrey,green,red,grey")


    def hooker(self):
        WM_KEYDOWN = 0x0100
        VK_DIVIDE = 0x6F
        def OnKeyboardEvent(event):
            if event.Key == 'Oem_7':
                user32.SetFocus(self.hwnd)
                user32.PostMessageW(self.hwnd, WM_KEYDOWN, VK_DIVIDE, 0)
                return False
            return True
        hm = pyHook.HookManager()
        hm.KeyDown = OnKeyboardEvent
        hm.HookKeyboard()


if __name__ == "__main__":
    app = scrabble_timer(None)
    app.mainloop()
